__name__ = "magico"
__version__ = "0.2"
__description__ = "MagicO - enabling attribute notation and JSONPath"
__url__ = "https://github.com/jackyko8/magico"
__author__ = "Jacky Ko"
__author_email__ = ""
__license__ = "MIT"
__copyright__ = "Copyright Jacky Ko"
